import sys
import threading

sys.path.append('../')
import shutil
from osgeo import gdal, osr
import numpy as np
import matplotlib.pyplot as plt
from random import randint
import os
from jacksung.utils.data_convert import nc2np, np2tif
from jacksung.utils.image import crop_png, zoom_image, zoomAndDock
from jacksung.utils.cache import Cache
import rasterio
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
from matplotlib.colors import LinearSegmentedColormap
import cartopy.feature as cfeature
import cartopy.crs as ccrs
from datetime import datetime, timedelta
import netCDF4 as nc
from jacksung.utils.data_convert import np2tif, Coordinate
from tqdm import tqdm
from jacksung.utils.multi_task import MultiTasks, type_process
import cv2

reference_cache = Cache(10)

x_range = {'left': -60, 'top': 60, 'bottom': -60, 'right': 60, 'width': 2400, 'height': 2400}
min_x_range = {'left': -60, 'top': 60, 'bottom': -60, 'right': 60, 'width': 480, 'height': 480}
static_params = {4000: {'l': 2747, 'c': 2747, 'COFF': 1373.5, 'CFAC': 10233137, 'LOFF': 1373.5, 'LFAC': 10233137},
                 2000: {'l': 5495, 'c': 5495, 'COFF': 2747.5, 'CFAC': 20466274, 'LOFF': 2747.5, 'LFAC': 20466274},
                 1000: {'l': 10991, 'c': 10991, 'COFF': 5495.5, 'CFAC': 40932549, 'LOFF': 5495.5, 'LFAC': 40932549},
                 500: {'l': 21983, 'c': 21983, 'COFF': 10991.5, 'CFAC': 81865099, 'LOFF': 10991.5, 'LFAC': 81865099},
                 250: {'l': 43967, 'c': 43967, 'COFF': 21983.5, 'CFAC': 163730199, 'LOFF': 21983.5, 'LFAC': 163730199}}


def getFY_coord(ld):
    return Coordinate(left=ld + x_range['left'], top=x_range['top'], right=ld + x_range['right'],
                      bottom=x_range['bottom'], h=x_range['height'], w=x_range['width'])
    # return Coordinate(left=ld - 45, top=36, right=ld + 45, bottom=-36, h=1571, w=1963)


def getFY_coord_min(ld):
    return Coordinate(left=ld + min_x_range['left'], top=min_x_range['top'], right=ld + min_x_range['right'],
                      bottom=min_x_range['bottom'], h=min_x_range['height'], w=min_x_range['width'])


def getFY_coord_clip(area=((100, 140, 10), (20, 60, 10))):
    return Coordinate(left=area[0][0], top=area[1][1], right=area[0][1], bottom=area[1][0], h=800, w=800)


# FY4星下点行列号转经纬度
def xy2coordinate(l, c, ld=105, res=4000):
    ea = 6378.137
    eb = 6356.7523
    h = 42164
    # 4000m分辨率
    COFF = static_params[res]['COFF']
    CFAC = static_params[res]['CFAC']
    LOFF = static_params[res]['LOFF']
    LFAC = static_params[res]['LFAC']

    x = (np.pi * (c - COFF)) / (180 * (2 ** -16) * CFAC)
    y = (np.pi * (l - LOFF)) / (180 * (2 ** -16) * LFAC)

    sd_t1 = np.square(h * np.cos(x) * np.cos(y))
    sd_t2 = np.square(np.cos(y)) + np.square(ea) / np.square(eb) * np.square(np.sin(y))
    sd_t3 = np.square(h) - np.square(ea)
    sd = np.sqrt(sd_t1 - sd_t2 * sd_t3)
    sn = (h * np.cos(x) * np.cos(y) - sd) / (np.cos(y) ** 2 + ea ** 2 / eb ** 2 * np.sin(y) ** 2)

    s1 = h - sn * np.cos(x) * np.cos(y)
    s2 = sn * np.sin(x) * np.cos(y)
    s3 = -sn * np.sin(y)
    sxy = np.sqrt(s1 ** 2 + s2 ** 2)

    lon = 180 / np.pi * np.arctan(s2 / s1) + ld
    lat = 180 / np.pi * np.arctan(ea ** 2 / eb ** 2 * s3 / sxy)
    return lon, lat


def convert_file2idx(file_name):
    file_name = file_name.replace('.npy', '')
    h, m = file_name[:2], file_name[2:4]
    return int(h) * 4 + int(m) // 15


def get_reference(ld):
    # 构造控制点列表 gcps_list
    gcps_list = []
    step = 50
    last_corrd = []
    lc_list = []
    latlon_list = []
    for l in range(0, 2748, step):
        for c in range(0, 2748, step):
            lon, lat = xy2coordinate(l, c, ld=ld)
            if str(lon) == 'nan' or str(lat) == 'nan':
                continue
            skip_flag = False
            for corrd in last_corrd:
                if (corrd[0] - lon) ** 2 + (corrd[1] - lat) ** 2 <= 100:
                    skip_flag = True
                    break
            if skip_flag:
                continue
            last_corrd.append([lon, lat])
            gcps_list.append(gdal.GCP(lon, lat, 0, c, l))
            lc_list.append((l, c))
            latlon_list.append((lon, lat))
    # 设置空间参考
    # print('控制点数目：', len(gcps_list))
    # print([(l, c) for l, c in lc_list])
    # print([(lon, lat) for lon, lat in latlon_list])
    spatial_reference = osr.SpatialReference()
    spatial_reference.SetWellKnownGeogCS('WGS84')
    return spatial_reference, gcps_list


def get_agri_file_path(data_path, data_date):
    e_date = data_date + timedelta(minutes=14, seconds=59)
    # /mnt/data1/szj/FY/downloaded_file
    # FY4B-_AGRI--_N_DISK_1330E_L1-_FDI-_MULT_NOM_20221007234500_20221007235959_4000M_V0001.HDF
    parent_dir = rf'{data_path}/{data_date.year}/{data_date.month}/{data_date.day}'
    start_date_str = data_date.strftime('%Y%m%d%H%M%S')
    end_date_str = e_date.strftime('%Y%m%d%H%M%S')
    file_path = rf'{parent_dir}/FY4B-_AGRI--_N_DISK_{get_ld(data_date)}0E_L1-_FDI-_MULT_NOM_{start_date_str}_{end_date_str}_4000M_V0001.HDF'
    if os.path.exists(file_path):
        return file_path
    else:
        return None


def getNPfromHDFClip(ld, file_path, file_type='FDI', lock=None, area=((100, 140, 10), (20, 60, 10)), cache=None):
    lon_d = int((ld - (area[0][0] + area[0][1]) / 2) * 20)
    lat_d = int(((area[1][0] + area[1][1]) / 2) * 20)
    np_data = getNPfromHDF(file_path, file_type, lock, cache=cache)
    np_data = np_data[:, 800 - lat_d:1600 - lat_d, 800 - lon_d:1600 - lon_d]
    return np_data


def getNPfromHDF(hdf_path, file_type='FDI', lock=None, cache=None, return_coord=False, only_coord=False):
    prase_data = prase_filename(os.path.basename(hdf_path))
    ld = prase_data['position']
    coord = Coordinate(left=ld - 60, top=60, right=ld + 60, bottom=-60, x_res=0.05, y_res=0.05)
    if only_coord:
        return coord
    if cache is not None:
        n_data = cache.get_key_in_cache(hdf_path + file_type)
        if n_data is None:
            n_data = _getNPfromHDF(hdf_path, file_type=file_type, lock=lock)
            cache.add_key(hdf_path + file_type, 'None' if n_data is None else n_data)
        if type(n_data) == str and n_data == 'None':
            if return_coord:
                return None, None
            else:
                return None
        else:
            if return_coord:
                return n_data, coord
            else:
                return n_data
    else:
        return _getNPfromHDF(hdf_path, file_type=file_type, lock=lock), coord


def _getNPfromHDF(hdf_path, file_type='FDI', lock=None):
    file_name = hdf_path.split(os.sep)[-1]
    file_info = prase_filename(file_name)
    if lock:
        lock.acquire()
    try:
        ds = nc.Dataset(hdf_path)
    except:
        print(f'open {hdf_path} failed')
    if lock:
        lock.release()
    if file_type == 'FDI':
        f = ds.groups['Data']
        np_data = np.zeros((15, 2748, 2748), dtype=np.float32)
        for i in range(1, 16):
            s_i = '0' + str(i) if i < 10 else str(i)
            data = np.array(f[f'NOMChannel{s_i}'][:]).astype(np.float32)
            data[data > 10000] = np.nan
            np_data[i - 1] = data
        # np_data = np_data[6:15]
        in_out_idx = [6, 15]
    elif file_type == 'QPE':
        np_data = np.array(ds['Precipitation'][:]).astype(np.float32)
        np_data = np_data[np.newaxis, :]
        in_out_idx = [0, 1]
    else:
        np_data = None
        in_out_idx = None
        raise Exception(rf'file_type {file_type} err')
    ds.close()
    r = reference_cache.get_key_in_cache(file_info['position'])
    if r is None:
        print(f'get reference of {file_info["position"]}')
        r = reference_cache.add_key(file_info['position'], get_reference(ld=file_info['position']))

    np_data = _getNPfromHDF_worker(np_data, file_info['start'], r=r, ld=file_info['position'], to_file=False,
                                   in_out_idx=in_out_idx)
    return np_data


def _getNPfromHDF_worker(read_np_data, current_date, data_dir=None, ld=None, r=None, to_file=True, in_out_idx=(6, 15)):
    tmp_dir = 'make_temp' + str(randint(10000000, 99999999))
    file = current_date.strftime("%H%M") + '.npy'
    os.makedirs(tmp_dir, exist_ok=True)
    save_path = f'{current_date.year}{current_date.month}{current_date.day}{file.replace(".npy", "")}'
    np2tif(read_np_data, tmp_dir, save_path, print_log=False)
    in_idx, out_idx = in_out_idx
    np_data = np.zeros((out_idx - in_idx, x_range['height'], x_range['width']), dtype=np.float16)
    for i in range(in_idx, out_idx):
        out_path = f'{tmp_dir}/{save_path}-{i}-ctrl.tif'
        _registration(f'{tmp_dir}/{save_path}-{i}.tif', out_path, ld, *r)
        img = cv2.imread(out_path, -1)
        if np.isnan(img).any():
            shutil.rmtree(tmp_dir)
            return None
        np_data[i - in_idx] = img.astype(np.float16)
    # raise Exception('manual stop')
    shutil.rmtree(tmp_dir)
    if to_file:
        os.makedirs(f'{data_dir}/dataset/{current_date.year}/{current_date.month}/{current_date.day}', exist_ok=True)
        np.save(
            f'{data_dir}/dataset/{current_date.year}/{current_date.month}/{current_date.day}/{convert_file2idx(file)}',
            np_data)
    else:
        return np_data


# 解析FY文件的文件名
def prase_filename(filename):
    m_list = filename.replace('.HDF', '').split('_')
    # FY4B-_AGRI--_N_DISK_1050E_L1-_FDI-_MULT_NOM_20250606171500_20250606172959_4000M_V0001.HDF
    return {'satellite': m_list[0], 'sensor': m_list[1], 'area': m_list[3], 'position': round(float(m_list[4][:3]), 2),
            'file_level': m_list[5], 'data_name': m_list[6], 'start': datetime.strptime(m_list[9], '%Y%m%d%H%M%S'),
            'end': datetime.strptime(m_list[10], '%Y%m%d%H%M%S'), 'resolution': m_list[11]}


def _registration(input_path, out_path, ld, spatial_reference, gcps_list):
    """
    基于python GDAL配准
    :param input_path: 需要配准的栅格文件
    :param out_path: 输出配准后的栅格文件位置
    :param top_left: 左上角坐标
    :param bottom_right: 右下角坐标
    :param ik: 行空白分辨率
    :param jk: 列空白分辨率
    :return:
    """
    # 打开栅格文件
    dataset = gdal.Open(input_path, gdal.GA_Update)
    # 添加控制点
    dataset.SetGCPs(gcps_list, spatial_reference.ExportToWkt())
    # tps校正 重采样:最邻近法
    gdal.Warp(out_path, dataset,
              format='GTiff',
              outputBounds=[ld + x_range['left'], x_range['bottom'], ld + x_range['right'], x_range['top']],
              resampleAlg=gdal.GRIORA_NearestNeighbour,
              width=x_range['width'],
              height=x_range['height'],
              tps=True,
              dstSRS='EPSG:4326')


def _prase_nc_worker(root_path, target_dir, file, lock=None):
    ps = prase_filename(file)
    file_name = ps["start"].strftime("%H%M")
    hdf_path = f'{root_path}/{target_dir}/{ps["start"].year}/{ps["start"].month}/{ps["start"].day}/{file}'
    save_path = f'{root_path}/npy/{ps["start"].year}/{ps["start"].month}/{ps["start"].day}'
    if not os.path.exists(hdf_path):
        tqdm.write(f'## {ps["start"].strftime("%Y%m%d %H:%M")} None ##')
        return False
    if os.path.exists(f'{save_path}/{convert_file2idx(file_name)}.npy'):
        tqdm.write(f'## {ps["start"].strftime("%Y%m%d %H:%M")} already exist ##')
        return True
    try:
        n_data = getNPfromHDF(hdf_path, lock=lock)
        os.makedirs(save_path, exist_ok=True)
        if n_data is None:
            tqdm.write(f'## {ps["start"].strftime("%Y%m%d %H:%M")} None ##')
            return False
            # raise Exception(f'nan in cliped {hdf_path}')
        np.save(f'{save_path}/{convert_file2idx(file_name)}', n_data)
    except Exception as e:
        # raise e
        tqdm.write(f'## {ps["start"].strftime("%Y%m%d %H:%M")} err ##')
        return False
    tqdm.write(f'{ps["start"].strftime("%Y%m%d %H:%M")} down')
    return True


# 把FY4的HDF文件转为npy文件
def make_fynp(root_path, target_dir, time_set):
    mt = MultiTasks(40)
    for current_date in time_set:
        for file in os.listdir(
                f'{root_path}/{target_dir}/{current_date.year}/{current_date.month}/{current_date.day}'):
            if file.endswith('.HDF'):
                mt.add_task(file, _prase_nc_worker, [root_path, target_dir, file, mt.thread_mutex])
                # _prase_nc_worker(root_path, target_dir, file, None)
    err_list = []
    for key, flag in mt.execute_task().items():
        if not flag:
            err_list.append(key + '\n')
    with open(f'err.log', 'w') as f:
        f.writelines(err_list)
    print(f'all done, {len(err_list)} in err.log')


def get_ld(data):
    if data < datetime(2024, 3, 1):
        return 133
    else:
        return 105


# 根据日期获取星下点位置
def get_filename_by_date(file_date):
    ld = get_ld(file_date)
    filename = rf'FY4B-_AGRI--_N_DISK_{ld}E_L1-_FDI-_MULT_NOM_{file_date.strftime("%Y%m%d%H%M%S")}_{(file_date + timedelta(minutes=14, seconds=59)).strftime("%Y%m%d%H%M%S")}_4000M_V0001.HDF'
    return filename


if __name__ == '__main__':
    # getNPfromHDF(rf'FY4B-_AGRI--_N_DISK_1330E_L2-_QPE-_MULT_NOM_20220702121500_20220702122959_4000M_V0001.NC',
    #              file_type='QPE')
    np_data = getNPfromHDF(
        rf'FY4B-_AGRI--_N_DISK_1330E_L1-_FDI-_MULT_NOM_20221230030000_20221230031459_4000M_V0001.HDF')
    np2tif(np_data, save_path='.', out_name='FY4B', left=133 - 60, top=60, x_res=0.05, y_res=0.05, dtype=np.float32)



import sys
import math
import cv2
import shutil
from jacksung.utils.multi_task import ThreadingLock
from PIL import ImageFont
from osgeo import gdal, osr
import numpy as np
import matplotlib

matplotlib.use('Agg')
import matplotlib.pyplot as plt
from random import randint
import os
from jacksung.utils.data_convert import nc2np, np2tif
from jacksung.utils.image import crop_png, zoom_image, zoomAndDock, draw_text, concatenate_images, make_block
from jacksung.utils.cache import Cache
import rasterio
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
from matplotlib.colors import LinearSegmentedColormap
import cartopy.feature as cfeature
import cartopy.crs as ccrs
import cartopy.io.img_tiles as cimgt
from rasterio.transform import from_origin
import yaml
import argparse
import jacksung.utils.fastnumpy as fnp
from tqdm import tqdm
from datetime import datetime, timedelta
from matplotlib.ticker import MaxNLocator
import netCDF4 as nc
import math
from jacksung.utils.data_convert import np2tif, Coordinate


def _get_color_normalization(data, colors):
    max_value = colors[-1][0]
    min_value = colors[0][0]
    data[data < min_value] = min_value
    data[data > max_value] = max_value
    data = (data - min_value) / (max_value - min_value)
    new_colors = []
    for idx, color in enumerate(colors):
        if idx == 0:
            value = 0
        elif idx == len(colors) - 1:
            value = 1
        else:
            value = (color[0] - min_value) / (max_value - min_value)
        new_colors.append([value, color[1]])
    return data, new_colors


# 色带颜色定位
def _get_color_position(value, colors):
    colors_min, colors_max = colors[0][0], colors[-1][0]
    colors = [[(color[0] - colors_min) / (colors_max - colors_min), color[1]] for color in colors]
    i = 0
    while i < len(colors) - 1:
        if value <= colors[i + 1][0]:
            break
        i += 1
    color_str0, color_str1 = colors[i][1], colors[i + 1][1]
    r1, g1, b1, r2, g2, b2 = int(color_str0[1:3], 16), int(color_str0[3:5], 16), int(color_str0[5:7], 16), \
        int(color_str1[1:3], 16), int(color_str1[3:5], 16), int(color_str1[5:7], 16)
    r = (value - colors[i][0]) / (colors[i + 1][0] - colors[i][0]) * (r2 - r1) + r1
    g = (value - colors[i][0]) / (colors[i + 1][0] - colors[i][0]) * (g2 - g1) + g1
    b = (value - colors[i][0]) / (colors[i + 1][0] - colors[i][0]) * (b2 - b1) + b1
    return np.array((b, g, r))


def make_color_map(colors, h, w, unit='', l_margin=300, r_margin=200, font_size=150, round_digits=1):
    colors_map = np.zeros((h, w, 3), dtype=np.uint8) + 255
    w = w - l_margin - r_margin
    for i in range(l_margin, w + l_margin):
        i = i - l_margin
        colors_map[:h - 150, i + l_margin] = _get_color_position(i / w, colors)
        if i in [0, w // 2, w - 1]:
            text_value = round((i / w) * (colors[-1][0] - colors[0][0]) + colors[0][0], round_digits)
            if round_digits == 0:
                text_value = int(text_value)
            text = str(text_value)
            if i == 0:
                text += unit
            colors_map = draw_text(colors_map, (i - 100 + l_margin, h - 150),
                                   font=ImageFont.truetype(
                                       rf'{os.path.abspath(os.path.dirname(__file__))}/../libs/times.ttf', font_size),
                                   text=text)
    return colors_map


def _make_fig(file_np,
              # [经度起,经度止,经度步长],[纬度起,纬度止,纬度步长]
              # np数据会自动填充整个图形,确保数据范围和area范围一致
              area, file_title='', save_name='img1.png',
              # 色带范围,请给出实际的数据范围
              # color=((0, '#1E90FF'), (2, '#1874CD'), (5, '#3A5FCD'), (10, '#0000CD'), (30, '#9400D3')),
              colors=None,
              colors_only=None,
              # 字体大小
              font_size=15,
              # 放大区域
              zoom_rectangle=(310 * 5, 300 * 5, 50 * 5, 40 * 5),
              # 放大区域停靠位置
              zoom_docker=(300, 730),
              # 图片清晰度
              dpi=500,
              xy_axis=None,
              draw_x_label=True,
              draw_y_label=True,
              draw_lon_grid=True,
              draw_lat_grid=True,
              clip_left=None,
              clip_right=None,
              clip_top=None,
              clip_bottom=None,
              # 添加各种特征
              # 自然海岸界,其他自带要素的参考cartopy
              # '10m', '50m', or '110m'
              features=(
                      cfeature.NaturalEarthFeature('physical', 'land', '50m', edgecolor='black', facecolor='none',
                                                   linewidth=0.4), cfeature.OCEAN, cfeature.LAND, cfeature.RIVERS),
              border_type=None, make_fig_lock=None):
    # corp = [92, 31, 542, 456]
    if xy_axis is None:
        xy_axis = area
    os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"
    # 设置经纬度范围,限定为中国
    # 注意指定crs关键字,否则范围不一定完全准确
    extents = [area[0][0], area[0][1], area[1][0], area[1][1]]
    if area[0][1] > 180:
        central_longitude = 180
    else:
        central_longitude = 0
    proj = ccrs.PlateCarree(central_longitude=central_longitude)
    data_crs = ccrs.PlateCarree()
    if make_fig_lock is not None:
        make_fig_lock.acquire()
    fig = plt.figure(dpi=dpi)
    ax = fig.add_subplot(111, projection=proj)

    if colors is None:
        np_min, np_max = np.nanpercentile(file_np, [5, 95])
        if colors_only is None:
            colors_only = ('#1E90FF', '#1874CD', '#3A5FCD', '#0000CD', '#9400D3')
        break_value = (np_max - np_min) / (len(colors_only) - 1)
        colors = [(np_min, colors_only[0])]
        for i in range(1, len(colors_only)):
            colors.append((np_min + i * break_value, colors_only[i]))
        colors.append((np_max, colors_only[-1]))
    # 用色带给数据上色,输入单通道,返回三通道图
    data_np, new_colors = _get_color_normalization(file_np, colors)
    cmap = LinearSegmentedColormap.from_list('custom_cmap', new_colors)
    for feature in features:
        ax.add_feature(feature)
    ax.imshow(data_np, origin='upper', extent=extents, transform=data_crs, cmap=cmap, vmin=new_colors[0][0],
              vmax=new_colors[-1][0])
    # 添加网格线
    if border_type is not None:
        # ax.gridlines(line_style='--')
        ax.gridlines(line_style=border_type)
    # 设置大刻度和小刻度
    tick_proj = ccrs.PlateCarree()
    if draw_x_label:
        ax.set_xticks(np.arange(xy_axis[0][0], xy_axis[0][1] + 1, xy_axis[0][2]), crs=tick_proj)
    # ax.set_xticks(np.arange(-180, 180 + 30, 30), minor=True, crs=tick_proj)
    if draw_y_label:
        ax.set_yticks(np.arange(xy_axis[1][0], xy_axis[1][1] + 1, xy_axis[1][2]), crs=tick_proj)
    # ax.set_yticks(np.arange(-90, 90 + 15, 15), minor=True, crs=tick_proj)
    # 利用Formatter格式化刻度标签
    if draw_lon_grid:
        ax.xaxis.set_major_formatter(LongitudeFormatter())
    if draw_lat_grid:
        ax.yaxis.set_major_formatter(LatitudeFormatter())
    ax.set_title(file_title, fontsize=font_size)
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    # plt.title(fontsize=font_size)
    # 关键修复：set_extent的crs应该是数据坐标系
    ax.set_extent(extents, crs=data_crs)
    plt.savefig(save_name)
    if zoom_rectangle is not None:
        read_png = cv2.imread(save_name)
        read_png = zoomAndDock(read_png, zoom_rectangle, zoom_docker, scale_factor=5, border=14)
        cv2.imwrite(save_name, read_png)
    np_data = cv2.imread(save_name) - 255
    np_sum_h = np.nonzero(np_data.sum(axis=(1, 2)))[0]
    np_sum_w = np.nonzero(np_data.sum(axis=(0, 2)))[0]
    # print(np_sum_h, np_sum_w)
    h, w = np_data.shape[:2]
    clip_left = min(np_sum_w[0], clip_left) if clip_left is not None else np_sum_w[0]
    clip_right = max(np_sum_w[-1], w - clip_right) if clip_right is not None else np_sum_w[-1]
    clip_top = min(np_sum_h[0], clip_top) if clip_top is not None else np_sum_h[0]
    clip_bottom = max(np_sum_h[-1], h - clip_bottom) if clip_bottom is not None else np_sum_h[-1]
    crop_png(save_name, left=clip_left, top=clip_top, right=clip_right, bottom=clip_bottom)
    plt.close()
    if make_fig_lock is not None:
        make_fig_lock.release()
    return colors
    # plt.show()


def make_fig(data,
             area,
             xy_axis=None,
             draw_x_label=True,
             draw_y_label=True,
             draw_lon_grid=True,
             draw_lat_grid=True,
             clip_left=None,
             clip_right=None,
             clip_top=None,
             clip_bottom=None,
             file_title='',
             save_name='figure_default.png',
             colors=None,
             colors_only=None,
             font_size=15,
             zoom_rectangle=None,
             zoom_docker=(300, 730),
             dpi=500,
             features=(
                     cfeature.NaturalEarthFeature('physical', 'land', '50m', edgecolor='black', facecolor='none',
                                                  linewidth=0.4), cfeature.OCEAN, cfeature.LAND, cfeature.RIVERS),
             border_type=None,
             draw_colormap=True,
             colormap_l_margin=300,
             colormap_r_margin=200,
             cm_font_size=150,
             round_digits=1,
             colormap_unit=''):
    colors = _make_fig(data, font_size=font_size, zoom_rectangle=zoom_rectangle, zoom_docker=zoom_docker, dpi=dpi,
                       features=features, border_type=border_type, xy_axis=xy_axis, draw_x_label=draw_x_label,
                       draw_y_label=draw_y_label, draw_lon_grid=draw_lon_grid, draw_lat_grid=draw_lat_grid,
                       file_title=file_title, save_name=save_name, area=area, colors=colors, colors_only=colors_only,
                       clip_left=clip_left, clip_right=clip_right, clip_top=clip_top, clip_bottom=clip_bottom)
    if draw_colormap:
        img = cv2.imread(save_name)
        h, w, c = img.shape
        cm = make_color_map(colors, 180, w, unit=colormap_unit, l_margin=colormap_l_margin, r_margin=colormap_r_margin,
                            font_size=cm_font_size, round_digits=round_digits)
        white_block = make_block(10, w)
        merge_img = concatenate_images([img, white_block, cm], direction='v')
        cv2.imwrite(save_name, merge_img)
    return colors


if __name__ == '__main__':
    data = np.load(r'C:\Users\ECNU\Desktop\delete_me\20230101_00_400_400.npy')[-1]
    data[data < 0.1] = np.nan
    make_fig(data, area=((40, 120, 20), (20, 60, 20)))



import os
import time
import pymysql
from datetime import datetime, timedelta
from tqdm import tqdm

import threading
import multiprocessing
import traceback, sys

import configparser


def convert_str(s):
    return "'" + str(s).replace("'", '"') + "'"


def convert_num(n):
    return str(n)


class BaseDB:
    def __init__(self, ini_path='db.ini'):
        self.conn = None
        self.db = None
        self.passwd = None
        self.user = None
        self.host = None
        self.port = None
        self.autocommit = True
        self.ini_path = ini_path
        pymysql.install_as_MySQLdb()
        self.lock_t = threading.Lock()

    def read_config(self):
        config = configparser.ConfigParser()
        config.read(self.ini_path)
        # Create the connection object
        self.host = config['database']['host']
        self.passwd = config['database']['password']
        self.db = config['database']['database']
        self.user = config['database'].get('user', 'root')
        self.port = config['database'].getint('port', 3306)

    def reconnect(self):
        if not self.conn:
            self.read_config()
        # 打开数据库连接
        connection = pymysql.connect(host=self.host, user=self.user, passwd=self.passwd, db=self.db, port=self.port,
                                     autocommit=self.autocommit)
        tqdm.write('Reconnected!')
        return connection

    def execute(self, sql):
        self.lock_t.acquire()
        cursor = None
        try:
            cursor = self.conn.cursor()
            result = cursor.execute(sql)
        except Exception as e:
            self.conn = self.reconnect()
            cursor = self.conn.cursor()
            try:
                result = cursor.execute(sql)
            except Exception as e:
                print(f'err SQL: {sql}')
                raise e
        finally:
            if cursor:
                cursor.close()
        self.lock_t.release()
        return result, cursor



import traceback

from satpy import Scene
from pyresample import create_area_def
import numpy as np
import os
from datetime import timedelta
from jacksung.utils.data_convert import np2tif, Coordinate
from contextlib import contextmanager


@contextmanager
def satpy_scene_context(filenames, reader):
    """satpy Scene上下文管理器"""
    scn = None
    try:
        scn = Scene(filenames=filenames, reader=reader)
        yield scn
    finally:
        # 显式清理satpy资源
        if scn is not None:
            try:
                # 尝试调用satpy的清理方法
                if hasattr(scn, 'close'):
                    scn.close()
                # 手动删除引用
                del scn
            except Exception as e:
                print(f"清理satpy资源时警告: {e}")


def _define_wgs84_area(resolution=0.05, area_extent=(-14.5, -60, 105.5, 60.0)):
    """定义WGS84坐标系目标区域（60°N-60°S，全经度），保持您原有的区域定义"""

    wgs84_proj = "+proj=longlat +datum=WGS84 +ellps=WGS84 +no_defs"  # 旧版projection参数（PROJ4字符串）
    target_area = create_area_def(
        area_id="wgs84_60n60s",
        projection=wgs84_proj,  # 旧版参数：projection（PROJ4字符串）
        area_extent=area_extent,  # 保持您原有的区域范围
        resolution=resolution,
        description=f"WGS84 Lat/Lon, 60N-60S, {resolution}deg resolution"
    )
    return target_area


def _extract_time_from_filename(filename):
    """从MSG文件名中提取时间戳（格式：20251126062741 → 2025-11-26_06-27-41）"""
    try:
        # MSG文件名格式示例：MSG2-SEVI-MSG15-0100-NA-20251126062741.272000000Z-NA.nat
        time_part = filename.split("-")[5].split(".")[0]  # 提取"20251126062741"
        formatted_time = f"{time_part[:4]}-{time_part[4:6]}-{time_part[6:8]}_{time_part[8:10]}-{time_part[10:12]}-{time_part[12:14]}"
        return formatted_time
    except IndexError:
        print("警告：无法从文件名提取时间戳，使用默认时间格式")
        return "unknown_time"


def _process_msg_seviri_to_numpy(nat_file_path, resolution=0.05, resampler="nearest", channels=("WV_062",), lock=None,
                                 only_ld=False):
    try:
        if lock is not None:
            lock.acquire()
        # ========== 主要改动开始 ==========
        # 使用上下文管理器确保Scene资源被释放（替换原来的Scene创建）
        with satpy_scene_context(filenames=[nat_file_path], reader="seviri_l1b_native") as scn:
            scn.load(channels)
            ld = scn[channels[0]].attrs['orbital_parameters']['projection_longitude']
            if only_ld:
                return ld
            area_extent = (ld - 60, -60, ld + 60, 60.0)
            target_area = _define_wgs84_area(resolution=resolution, area_extent=area_extent)
            scn_wgs84 = scn.resample(target_area, resampler=resampler)
            time_str = _extract_time_from_filename(os.path.basename(nat_file_path))
            result = {
                'data': {},
                'metadata': {},
                'coordinates': {},
                'global_attrs': {
                    'source_file': os.path.basename(nat_file_path),
                    'processing_time': time_str,
                    'resolution_degrees': resolution,
                    'resampling_method': resampler,
                    'area_extent': area_extent
                }
            }
            # 提取每个通道的数据
            for ch in channels:
                try:
                    data_array = scn_wgs84[ch].values
                    result['data'][ch] = data_array
                    result['metadata'][ch] = dict(scn_wgs84[ch].attrs)
                except Exception as e:
                    print(f"提取{ch}通道数据失败：{str(e)}")
            # 提取坐标信息
            if channels:
                first_ch = channels[0]
                try:
                    area = scn_wgs84[first_ch].attrs['area']
                    lons, lats = area.get_lonlats()
                    result['coordinates']['longitude'] = lons
                    result['coordinates']['latitude'] = lats
                    result['coordinates']['shape'] = lons.shape
                except Exception as e:
                    print(f"提取坐标信息失败：{str(e)}")
            return result
    except Exception as e:
        print(f"处理失败：{str(e)}")
        traceback.print_exc()
        return None
    finally:
        if lock is not None:
            lock.release()


def getNPfromNAT(file_path, save_file=False, lock=None, return_coord=False, only_coord=False):
    all_target_channels = ["WV_062", "WV_073", "IR_087", "IR_097", "IR_108", "IR_120", "IR_134"]
    # all_target_channels = ["VIS006", "VIS008"]
    if only_coord:
        ld = _process_msg_seviri_to_numpy(nat_file_path=file_path, resolution=0.05, resampler="nearest",
                                          channels=all_target_channels, lock=lock, only_ld=True)
        return Coordinate(left=ld - 60, bottom=-60, right=ld + 60, top=60, x_res=0.05, y_res=0.05)
    else:
        result = _process_msg_seviri_to_numpy(nat_file_path=file_path, resolution=0.05, resampler="nearest",
                                              channels=all_target_channels, lock=lock, only_ld=False)
    np_data = None
    coord = None
    if result is not None:
        for idx, channel in enumerate(all_target_channels):
            if channel in result['data']:
                chann_data = result['data'][channel].copy()
                if np_data is None:
                    np_data = np.zeros((len(all_target_channels),) + chann_data.shape, dtype=chann_data.dtype)
                    area_extent = result['global_attrs']['area_extent']
                    coord = Coordinate(left=area_extent[0], bottom=area_extent[1], right=area_extent[2],
                                       top=area_extent[3], x_res=0.05, y_res=0.05)
                np_data[idx] = chann_data
            else:
                del result
                raise Exception(f"文件{file_path}，通道 {channel} 数据未找到")
    else:
        raise Exception(f"文件{file_path}处理失败，未获取数据")
    if save_file:
        np2tif(np_data, save_path='np2tif_dir', coord=coord, out_name='MetSat',
               dtype='float32')
    if return_coord:
        return np_data, coord
    else:
        return np_data


def get_seviri_file_path(data_path, data_date):
    e_date = data_date + timedelta(minutes=14, seconds=59)
    parent_dir = rf'{data_path}/{data_date.year}/{data_date.month}/{data_date.day}/'
    start_date_str = data_date.strftime('%Y%m%d%H%M%S')
    end_date_str = e_date.strftime('%Y%m%d%H%M%S')
    for file in os.listdir(parent_dir):
        if file.endswith('.nat') and start_date_str <= file.split('-')[5].split('.')[0] <= end_date_str:
            return os.path.join(parent_dir, file)
    return None


if __name__ == '__main__':
    np_data = getNPfromNAT("MSG4-SEVI-MSG15-0100-NA-20221230031243.610000000Z-NA.nat", save_file=False)
    print()



import netCDF4 as nc
import numpy as np
from einops import rearrange, repeat
from jacksung.utils.data_convert import np2tif, get_transform_from_lonlat_matrices


def getNPfromHDF(hdf_path, lock=None, save_file=True):
    if lock:
        lock.acquire()
    ds = nc.Dataset(hdf_path)
    if lock:
        lock.release()
    np_data = np.array(ds['hourlyPrecipRateGC'][:]).astype(np.float32)[0]
    ds.close()
    np_data[np_data < 0] = np.nan
    np_data = np_data[::-1, :]
    if save_file:
        np2tif(np_data, save_path='np2tif_dir', out_name='gsmap', dtype='float32',
               left=-180, top=90, x_res=0.1, y_res=0.1)
    return np_data


if __name__ == '__main__':
    getNPfromHDF(rf'D:\python_Project\Huayu_Global\file_download\gsmap_now_rain.20220702.0300.nc')



import hashlib


def calculate_file_hash(file_path, hash_algorithm="md5", chunk_size=4096):
    hash_obj = hashlib.new(hash_algorithm)
    with open(file_path, "rb") as file:
        while True:
            data = file.read(chunk_size)
            if not data:
                break
            hash_obj.update(data)
    return hash_obj.hexdigest()


def hash_files(file_paths, hash_algorithm="md5"):
    hash_list = []
    for filepath in file_paths:
        hash_list.append(calculate_file_hash(filepath, hash_algorithm))
    return hash_string(''.join(hash_list), hash_algorithm)


def hash_string(s, hash_algorithm="md5"):
    hash_obj = hashlib.new(hash_algorithm)
    # 对连接后的字符串进行哈希
    hash_obj.update(s.encode())
    return hash_obj.hexdigest()



import os

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import math
from pytorch_msssim import ssim
from torchmetrics.image import PeakSignalNoiseRatio, StructuralSimilarityIndexMeasure
from torchmetrics import R2Score, PearsonCorrCoef, AUROC
from torchmetrics.regression import MeanSquaredError
import importlib
from einops import rearrange
import cv2


def compute_rmse(da_fc, da_true):
    error = da_fc - da_true
    error = error ** 2
    number = torch.sqrt(error.mean((-2, -1)))
    return number.mean()


class Metrics:
    def __init__(self):
        self.psnr = PeakSignalNoiseRatio()
        self.ssim = StructuralSimilarityIndexMeasure()
        self.rr = R2Score()
        self.p = PearsonCorrCoef()
        self.AUROC = AUROC("binary")

    def mask_nan(self, pred, target):
        pred = pred.flatten()
        target = target.flatten()
        # 生成有效掩码：真实值和预测值均非 NaN 的位置为 True
        valid_mask = ~(torch.isnan(pred) | torch.isnan(target))

        # 过滤无效元素
        pred = pred[valid_mask]
        target = target[valid_mask]

        return pred, target

    def calc_AUROC(self, preds, targets):
        AUROC = 0
        for i in range(len(preds)):
            pred, target = self.mask_nan(preds[i], targets[i])
            AUROC += self.AUROC(pred, target)
        self.AUROC.reset()
        return AUROC / len(preds)

    def calc_psnr(self, preds, targets):
        psnr = 0
        for i in range(len(preds)):
            psnr += self.psnr(rearrange(preds[i], '(b c) h w->b c h w', b=1),
                              rearrange(targets[i], '(b c) h w->b c h w', b=1))
        self.psnr.reset()
        return psnr / len(preds)

    def calc_ssim(self, preds, targets):
        ssim = 0
        for i in range(len(preds)):
            ssim += self.ssim(rearrange(preds[i], '(b c) h w->b c h w', b=1),
                              rearrange(targets[i], '(b c) h w->b c h w', b=1))
        self.ssim.reset()
        return ssim / len(preds)

    def calc_rmse(self, preds, targets):
        rmse = 0
        for i in range(len(preds)):
            rmse += compute_rmse(preds[i], targets[i])
        return rmse / len(preds)

    def calc_rr(self, preds, targets):
        rr = 0
        for i in range(len(preds)):
            pred, tar = self.mask_nan(preds[i], targets[i])
            rr += self.rr(pred, tar)
        self.rr.reset()
        return rr / len(preds)

    def calc_p(self, preds, targets, exclude_zero=False):
        p = 0
        count = 0
        for i in range(len(preds)):
            pred, target = self.mask_nan(preds[i], targets[i])
            if exclude_zero:
                mask = target != 0
                pred = pred[mask]
                target = target[mask]
            if pred.var() == 0 or target.var() == 0:
                continue
            count += 1
            p += self.p(pred, target)
        self.p.reset()
        return p / count

    def print_metrics(self, preds, targets, print_log=True):
        rr = float(self.calc_rr(preds, targets))
        p = float(self.calc_p(preds, targets))
        rmse = float(self.calc_rmse(preds, targets))
        ssim = float(self.calc_ssim(preds, targets))
        psnr = float(self.calc_psnr(preds, targets))
        if print_log:
            print(rf'p: {p} rr: {rr} rmse: {rmse} ssim: {ssim} psnr: {psnr}')
        return {'p': p, 'rr': rr, 'rmse': rmse, 'ssim': ssim, 'psnr': psnr}

    def calculate_rain_metrics(self, preds, targets, threshold=0.1):
        """
        使用flatten()批量计算多张降雨图的POD、FAR、ACC、CSI指标

        参数:
            preds: 预测的降雨tensor，形状为[样本数, 高度, 宽度]
            target: 观测的降雨tensor，形状与preds相同
            threshold: 降雨事件的阈值，默认0.1mm

        返回:
            metrics: 包含每个样本指标的字典
        """
        # 1. 将每个样本的空间维度展平（[样本数, 高度, 宽度] → [样本数, 像素总数]）
        preds_flat = preds.flatten(start_dim=1)  # 从第1维开始展平（保留样本维度）
        targets_flat = targets.flatten(start_dim=1)

        # 2. 二值化（1=有雨，0=无雨）
        preds_binary = (preds_flat >= threshold).float()
        targets_binary = (targets_flat >= threshold).float()

        # 3. 计算混淆矩阵元素（按样本维度求和）
        TP = torch.sum(preds_binary * targets_binary, dim=1)  # 每个样本的TP总和
        FP = torch.sum(preds_binary * (1 - targets_binary), dim=1)
        TN = torch.sum((1 - preds_binary) * (1 - targets_binary), dim=1)
        FN = torch.sum((1 - preds_binary) * targets_binary, dim=1)

        # 4. 计算指标（处理分母为0的情况）
        POD = TP / (TP + FN)
        FAR = FP / (TP + FP)
        ACC = (TP + TN) / (TP + FP + TN + FN)
        CSI = TP / (TP + FP + FN)

        # 5. 标记无效值为NaN
        POD = torch.where((TP + FN) == 0, torch.nan, POD)
        FAR = torch.where((TP + FP) == 0, torch.nan, FAR)
        ACC = torch.where((TP + FP + TN + FN) == 0, torch.nan, ACC)
        CSI = torch.where((TP + FP + FN) == 0, torch.nan, CSI)

        return POD, FAR, ACC, CSI


def img2tensor(img):
    if type(img) == str:
        img = cv2.imread(img, -1)
    img = torch.from_numpy(img)
    img = rearrange(img, ' (b c h) w->b c h w', b=1, c=1)
    return img


if __name__ == '__main__':
    preds = torch.Tensor([[[0, 0, 1],
                           [0, 1, 1],
                           [1, 1, 1]
                           ],
                          [[0, 1, 1],
                           [0, 1, 1],
                           [0, 1, 1]
                           ]])
    # target = torch.rand(2, 1, 3, 3)
    target = torch.Tensor([[[0, 1, 1],
                            [0, 1, 1],
                            [0, 1, 1]
                            ],
                           [[0, 1, 1],
                            [0, 1, 1],
                            [0, 0, 1]
                            ]])
    target[1, 0, 2] = torch.nan
    # m = Metrics()
    # print(m.calc_rr(preds, target))
    m = Metrics()
    AUROC = m.calc_AUROC(preds, target)
    print(AUROC)
    print(m.calculate_rain_metrics(preds, target))



import netCDF4 as nc
import numpy as np
from einops import rearrange, repeat
from jacksung.utils.data_convert import np2tif, get_transform_from_lonlat_matrices


def getNPfromHDF(hdf_path, lock=None, save_file=True, print_log=False):
    if lock:
        lock.acquire()
    ds = nc.Dataset(hdf_path)
    if lock:
        lock.release()
    np_data = np.array(ds['cmorph'][:]).astype(np.float32)
    lon_array = np.array(ds['lon'][:]).astype(np.float32)
    lat_array = np.array(ds['lat'][:]).astype(np.float32)
    lon_dim = len(lon_array)
    lat_dim = len(lat_array)
    lon_array = repeat(lon_array, 'w -> w h', h=lat_dim)
    lat_array = repeat(lat_array, 'h -> w h', w=lon_dim)
    ds.close()
    # np_data = rearrange(np_data[0], 'w h->h w')[::-1, :]
    np_data[np_data < 0] = 0
    # np_data = np_data[0] + np_data[1]
    transform, avg_error = get_transform_from_lonlat_matrices(
        lon_array=lon_array,
        lat_array=lat_array,
        gcp_density=20,  # 范围越大，gcp_density建议越大
        print_log=print_log
    )
    if save_file:
        np2tif(np_data, save_path='np2tif_dir', out_name='CMORPH', dtype='float32', transform=transform)
    return np_data, transform


if __name__ == '__main__':
    data = getNPfromHDF(rf'C:\Users\ECNU\PycharmProjects\CMORPH_V1.0_ADJ_8km-30min_2022070203.nc')
    # from datetime import datetime
    #
    # da = datetime.utcfromtimestamp(1656730800)
    # print(da)
    # da = datetime.utcfromtimestamp(1656732600)
    # print(da)



from torch.nn.parallel import DataParallel
import torch
from torch.nn.parallel._functions import Scatter
from torch.nn.parallel.parallel_apply import parallel_apply


def scatter(inputs, target_gpus, chunk_sizes, dim=0):
    r"""
    Slices tensors into approximately equal chunks and
    distributes them across given GPUs. Duplicates
    references to objects that are not tensors.
    """

    def scatter_map(obj):
        if isinstance(obj, torch.Tensor):
            try:
                return Scatter.apply(target_gpus, chunk_sizes, dim, obj)
            except:
                print('obj', obj.size())
                print('dim', dim)
                print('chunk_sizes', chunk_sizes)
                quit()
        if isinstance(obj, tuple) and len(obj) > 0:
            return list(zip(*map(scatter_map, obj)))
        if isinstance(obj, list) and len(obj) > 0:
            return list(map(list, zip(*map(scatter_map, obj))))
        if isinstance(obj, dict) and len(obj) > 0:
            return list(map(type(obj), zip(*map(scatter_map, obj.items()))))
        return [obj for targets in target_gpus]

    # After scatter_map is called, a scatter_map cell will exist. This cell
    # has a reference to the actual function scatter_map, which has references
    # to a closure that has a reference to the scatter_map cell (because the
    # fn is recursive). To avoid this reference cycle, we set the function to
    # None, clearing the cell
    try:
        return scatter_map(inputs)
    finally:
        scatter_map = None


def scatter_kwargs(inputs, kwargs, target_gpus, chunk_sizes, dim=0):
    r"""Scatter with support for kwargs dictionary"""
    inputs = scatter(inputs, target_gpus, chunk_sizes, dim) if inputs else []
    kwargs = scatter(kwargs, target_gpus, chunk_sizes, dim) if kwargs else []
    if len(inputs) < len(kwargs):
        inputs.extend([() for _ in range(len(kwargs) - len(inputs))])
    elif len(kwargs) < len(inputs):
        kwargs.extend([{} for _ in range(len(inputs) - len(kwargs))])
    inputs = tuple(inputs)
    kwargs = tuple(kwargs)
    return inputs, kwargs


class BalancedDataParallel(DataParallel):
    def __init__(self, gpu0_bsz, *args, **kwargs):
        self.gpu0_bsz = gpu0_bsz
        super().__init__(*args, **kwargs)

    def forward(self, *inputs, **kwargs):
        if not self.device_ids:
            return self.module(*inputs, **kwargs)
        gpu_idx_start = 1 if self.gpu0_bsz == 0 else 0
        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids[gpu_idx_start:])
        if len(self.device_ids) == 1:
            return self.module(*inputs[0], **kwargs[0])
        replicas = self.replicate(self.module, self.device_ids[:len(inputs) + gpu_idx_start])
        replicas = replicas[gpu_idx_start:]
        outputs = self.parallel_apply(replicas, self.device_ids[gpu_idx_start:], inputs, kwargs)
        return self.gather(outputs, self.output_device)

    def parallel_apply(self, replicas, device_ids, inputs, kwargs):
        return parallel_apply(replicas, inputs, kwargs, device_ids[:len(inputs)])

    def scatter(self, inputs, kwargs, device_ids):
        bsz = inputs[0].size(self.dim)
        num_dev = len(self.device_ids)
        gpu0_bsz = self.gpu0_bsz
        bsz_unit = (bsz - gpu0_bsz) // (num_dev - 1)
        if gpu0_bsz < bsz_unit:
            chunk_sizes = [gpu0_bsz] + [bsz_unit] * (num_dev - 1)
            delta = bsz - sum(chunk_sizes)
            for i in range(delta):
                chunk_sizes[i + 1] += 1
            if gpu0_bsz == 0:
                chunk_sizes = chunk_sizes[1:]
        else:
            return super().scatter(inputs, kwargs, device_ids)

        return scatter_kwargs(inputs, kwargs, device_ids, chunk_sizes, dim=self.dim)



import numpy as np
import requests
from urllib.parse import quote
import _thread
import time
import threading
import sys

threadLock = threading.Lock()


def format_log(*args):
    return '[' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + '] ' + ' '.join([str(x) for x in args])


def oprint(*args, **kwargs):
    log = format_log(*args)
    print(log, end=kwargs.get('end', '\n'), flush=kwargs.get('flush', False))


def thread_send_log(url, content, name):
    threadLock.acquire()
    content = quote(content, 'utf-8')
    name = quote(name, 'utf-8')
    url = url + '&name=' + name + '&content=' + content
    # print('sendLog:' + url)
    try:
        # print("----------------sendLog...----------------")
        requests.get(url, timeout=5)
        # print('\nsendLog finish', r.status_code, r.content)
        # print('sendLog finish')
    except Exception as e:
        print('\nsendLog network error!')
    finally:
        # print("----------------sendLog...----------------")
        threadLock.release()


class StdLog(object):
    def __init__(self, filename='default.log', common_path='warning_log', stream=sys.stdout):
        self.terminal = stream
        self.log = open(filename, 'a')
        self.common_log = None
        self.common_path = common_path

    def write(self, message):
        message = str(message)
        if message.count('[TemporaryTag]') == 0:
            if message.count('[Common]') != 0 or message.count('[Warning]') != 0 \
                    or message.count('[Error]') != 0 or message.count('[OnlyFile]') != 0:
                if self.common_log is None:
                    self.common_log = open(self.common_path, 'a')
                self.common_log.write(message.replace('[Common]', '').replace('[OnlyFile]', ''))
                self.common_log.flush()
            else:
                self.log.write(message)
                self.log.flush()
        else:
            message = message.replace('[TemporaryTag]', '')
        if message.count('[OnlyFile]') == 0:
            self.terminal.write(message)

    def flush(self):
        self.terminal.flush()
        self.log.flush()
        if self.common_log is not None:
            self.common_log.flush()


class LogClass:
    def __init__(self, on=False, url=None):
        self.on = on
        self.url = url

    def send_log(self, content, name):
        if self.on:
            try:
                _thread.start_new_thread(thread_send_log, (self.url, content, name))
            except Exception as e:
                print("Cloud Log Error")


if __name__ == '__main__':
    # sys.stdout = StdLog(filename='log.txt', common_path='warning.txt')
    oprint('this is a log', end=' ')
    oprint('this is a log')



import torch.nn as nn
from jacksung.ai.GeoNet.m_blockV2 import FEB, Tail, Head, DownBlock, UpBlock, CubeEmbeding, CubeUnEmbeding, Norm, ACT
import torch
import jacksung.utils.fastnumpy as fnp


class GeoNet(nn.Module):
    def __init__(self, window_sizes, n_lgab, c_in, c_lgan, r_expand=4, down_sample=2, num_heads=8, downstage=2):
        super(GeoNet, self).__init__()
        self.window_sizes = window_sizes
        self.n_lgab = n_lgab
        self.c_in = c_in
        self.c_lgan = c_lgan
        self.r_expand = r_expand
        self.down_sample = down_sample
        # define head module
        self.head = Head(self.c_lgan, self.c_in, self.down_sample)
        # define body module
        self.body = nn.ModuleList()
        self.downstage = downstage
        for i in range(self.n_lgab):
            if i % self.downstage in [1]:
                if 0 <= i < self.n_lgab / 2 - 1:
                    self.body.append(DownBlock(self.c_lgan, 2))
                    self.c_lgan = self.c_lgan * 2
                elif i > self.n_lgab / 2 + 1:
                    self.body.append(UpBlock(self.c_lgan // 2, 2))
                    self.c_lgan = self.c_lgan // 2
            self.body.append(
                FEB(self.c_lgan, self.r_expand, self.window_sizes[i % len(self.window_sizes)], num_heads=num_heads))
        self.tail = Tail(self.c_lgan, 3, down_sample)

    def forward(self, x, roll=0):
        # head

        x = torch.roll(x, shifts=roll, dims=-1)
        x = self.head(x)
        x_res = list()
        x_res.append(x)
        idx = 0
        for stage in self.body:
            if str(stage.__class__) == "<class 'jacksung.ai.GeoNet.m_blockV2.FEB'>":
                # 7,9
                if idx > self.n_lgab / 2 + 1 and idx % self.downstage in [1]:
                    x += x_res.pop()
                # 0,2
                x = stage(x)
                if idx < self.n_lgab / 2 - 1 and idx % self.downstage in [0]:
                    x_res.append(x)
                idx += 1
            else:
                x = stage(x)
        # tail
        x += x_res.pop()
        x = self.tail(x)
        if roll > 0:
            x = torch.roll(x, shifts=-roll, dims=-1)
        return x

    def init_model(self):
        print('Initializing the model!')
        for m in self.children():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.xavier_normal_(m.weight)

    def load(self, state_dict, strict=True):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            name = name[name.index('.') + 1:]
            if name in own_state.keys():
                if isinstance(param, nn.Parameter):
                    param = param.data
                try:
                    own_state[name].copy_(param)
                    # own_state[name].requires_grad = False
                except Exception as e:
                    err_log = f'While copying the parameter named {name}, ' \
                              f'whose dimensions in the model are {own_state[name].size()} and ' \
                              f'whose dimensions in the checkpoint are {param.size()}.'
                    if not strict:
                        print(err_log)
                    else:
                        raise Exception(err_log)
            elif strict:
                raise KeyError(f'unexpected key {name} in {own_state.keys()}')
            else:
                print(f'{name} not loaded by model')


if __name__ == '__main__':
    pass



import torch
import torch.nn as nn
import cv2
import numpy as np
import matplotlib.pyplot as plt
import os
from jacksung.utils.image import crop_png, zoom_image, zoomAndDock
import rasterio
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
from matplotlib.colors import LinearSegmentedColormap
import cartopy.feature as cfeature
import cartopy.crs as ccrs
import yaml
import argparse
from datetime import datetime, timedelta
from matplotlib.ticker import MaxNLocator
import random


def load_model(model, state_dict, strict=True):
    own_state = model.state_dict()
    for name, param in state_dict.items():
        name = name[name.index('.') + 1:]
        if name in own_state.keys():
            if isinstance(param, nn.Parameter):
                param = param.data
            try:
                own_state[name].copy_(param)
                # own_state[name].requires_grad = False
            except Exception as e:
                err_log = f'While copying the parameter named {name}, ' \
                          f'whose dimensions in the model are {own_state[name].size()} and ' \
                          f'whose dimensions in the checkpoint are {param.size()}.'
                if not strict:
                    print(err_log)
                else:
                    raise Exception(err_log)
        elif strict:
            raise KeyError(f'unexpected key {name} in {own_state.keys()}')
        else:
            print(f'{name} not loaded by model')


def get_stat_dict(metrics, extra_info=None):
    if extra_info is None:
        extra_info = dict()
    stat_dict = {'epochs': 0, 'loss': [], 'metrics': {}}
    for idx, metrics in enumerate(metrics):
        name, default_value, op = metrics
        stat_dict['metrics'][name] = {'value': [], 'best': {'value': default_value, 'epoch': 0, 'op': op}}
    for key, value in extra_info.items():
        stat_dict[key] = value
    return stat_dict


def data_to_device(datas, device, fp=32):
    outs = []
    for data in datas:
        if fp == 16:
            data = data.type(torch.HalfTensor)
        if fp == 64:
            data = data.type(torch.DoubleTensor)
        if fp == 32:
            data = data.type(torch.FloatTensor)
        data = data.to(device)
        outs.append(data)
    return outs


def draw_lines(stat_dict_path):
    os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'
    print('[TemporaryTag]Producing the LinePicture of Log...', end='[TemporaryTag]\n')
    yaml_args = yaml.load(open(stat_dict_path), Loader=yaml.FullLoader)
    # 创建图表
    m_len = len(yaml_args['metrics'])
    sub_loss_count = 0
    for i in range(0, 5):
        if f'loss{i}es' in yaml_args:
            sub_loss_count += 1
        else:
            break
    plt.figure(figsize=(8 * (m_len + sub_loss_count), 6))  # 设置图表的大小
    x = np.array(range(1, yaml_args['epochs'] + 1))
    for idx, d in enumerate(yaml_args['metrics'].items()):
        m_name, m_value = d
        y = np.array(m_value['value'])
        # 生成数据
        plt.subplot(1, m_len + sub_loss_count, idx + 1)
        plt.plot(x, y)
        plt.title(m_name)
        plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
    for i in range(0, sub_loss_count):
        y = np.array(yaml_args[f'loss{i}es'])
        plt.subplot(1, m_len + sub_loss_count, m_len + i + 1)
        scale = len(y) / yaml_args['epochs']
        # 把X坐标的纯统计范围缩放到和其他图表一致(epoch)
        x = np.array(range(1, len(y) + 1)) / scale
        plt.plot(x, y)
        plt.title(f'Loss{i}')
        plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
    # 添加图例
    # plt.legend()
    plt.savefig(os.path.join(os.path.dirname(stat_dict_path), 'Metrics.jpg'))


def make_best_metric(stat_dict, metrics, epoch, save_model_param, server_log_param):
    save_model_flag = False
    experiment_model_path, model, optimizer, scheduler = save_model_param
    log, epochs, cloudLogName = server_log_param

    for name, m_value in metrics:
        stat_dict['metrics'][name]['value'].append(m_value)
        inf = float('inf')
        if eval(str(m_value) + stat_dict['metrics'][name]['best']['op'] + str(
                stat_dict['metrics'][name]['best']['value'])):
            stat_dict['metrics'][name]['best']['value'] = m_value
            stat_dict['metrics'][name]['best']['epoch'] = epoch
            log.send_log('{}:{} epoch:{}/{}'.format(name, m_value, epoch, epochs), cloudLogName)
            save_model_flag = True

    if save_model_flag:
        # sava best model
        save_model(os.path.join(experiment_model_path, 'model_{}.pt'.format(epoch)), epoch,
                   model, optimizer, scheduler, stat_dict)
    # '[Validation] nRMSE/RMSE: {:.4f}/{:.4f} (Best: {:.4f}/{:.4f}, Epoch: {}/{})\n'
    test_log = f'[Val epoch:{epoch}] ' \
               + ' '.join([str(m[0]) + ':' + str(round(m[1], 4)) + '('
                           + str(round(stat_dict['metrics'][m[0]]['best']['value'], 4)) + ')' for m in metrics]) \
               + ' (Best Epoch: ' \
               + '/'.join([str(stat_dict['metrics'][m[0]]['best']['epoch']) for m in metrics]) \
               + ')'
    save_model(os.path.join(experiment_model_path, 'model_latest.pt'), epoch, model, optimizer, scheduler, stat_dict)
    return test_log


def save_model(_path, _epoch, _model, _optimizer, _scheduler, _stat_dict):
    # torch.save(model.state_dict(), saved_model_path)
    torch.save({
        'epoch': _epoch,
        'model_state_dict': _model.state_dict(),
        'optimizer_state_dict': _optimizer.state_dict(),
        'scheduler_state_dict': _scheduler.state_dict(),
        'stat_dict': _stat_dict
    }, _path)


def parse_config(config=None, set_gpu=True):
    parser = argparse.ArgumentParser(description='config')
    parser.add_argument('--config', type=str, default=None, help='pre-config file for training')
    parser.add_argument('--prec_data_path', type=str, default=None, help='dataset path')
    args = parser.parse_args()
    if args.config:
        opt = vars(args)
        yaml_args = yaml.load(open(args.config), Loader=yaml.FullLoader)
        opt.update(yaml_args)
    else:
        opt = vars(args)
        yaml_args = yaml.load(open(config), Loader=yaml.FullLoader)
        opt.update(yaml_args)

    if set_gpu:
        # set visible gpu
        os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
        os.environ['CUDA_VISIBLE_DEVICES'] = ','.join([str(x) for x in args.gpu_ids])

        # select active gpu devices
        if args.gpu_ids is not None and torch.cuda.is_available():
            print('use cuda & cudnn for acceleration!')
            print('the gpu id is: {}'.format(args.gpu_ids))
            device = torch.device('cuda')
            # device = torch.device('cuda:' + str(args.gpu_ids[0]))
        else:
            print('use cpu for training!')
            device = torch.device('cpu')
        return device, args
    else:
        return args


def _get_color_normalization(data, colors):
    max_value = colors[-1][0]
    min_value = colors[0][0]
    data[data < min_value] = min_value
    data[data > max_value] = max_value
    data = (data - min_value) / (max_value - min_value)
    new_colors = []
    for color in colors:
        new_colors.append([(color[0] - min_value) / (max_value - min_value), color[1]])
    return data, new_colors


def data_augmentation(images):
    # 旋转
    rotate = random.random()
    if 0 <= rotate < 0.25:
        images = [torch.rot90(image, 1, [2, 3]) for image in images]
    elif 0.25 <= rotate < 0.5:
        images = [torch.rot90(image, 2, [2, 3]) for image in images]
    elif 0.5 <= rotate < 0.75:
        images = [torch.rot90(image, 3, [2, 3]) for image in images]
    # 水平翻折
    if random.random() > 0.5:
        images = [torch.flip(image, [2]) for image in images]
    # 垂直翻折
    if random.random() > 0.5:
        images = [torch.flip(image, [3]) for image in images]
    return images


def make_fig(file_name, root_path, out_folder=None, tz='UTC',
             colors=((0, '#1E90FF'), (0.1, '#1874CD'), (0.2, '#3A5FCD'), (0.3, '#0000CD'), (1, '#9400D3')),
             area=((100, 140, 10), (20, 60, 10)), font_size=20, corp=(0, 0, None, None),
             zoom_rectangle=(310 * 5, 300 * 5, 50 * 5, 40 * 5), docker=(300, 730), dpi=500, filter=0.3, exposure=None):
    # corp = [92, 31, 542, 456]
    print('过期的make_figure方法！')
    os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"
    extents = [100, 140, 20, 60]
    proj = ccrs.PlateCarree()
    fig = plt.figure(dpi=dpi)
    ax = fig.add_subplot(111, projection=proj)
    ax.set_extent(extents, crs=proj)
    # 读取TIFF数据
    elevation = None
    if type(file_name) == list:
        for each_file in file_name:
            file_path = os.path.join(root_path, each_file)
            with rasterio.open(file_path) as dataset:
                el_rd = dataset.read(1)
                elevation[elevation < filter] = np.nan
                if elevation is None:
                    elevation = el_rd
                else:
                    elevation += el_rd
    else:
        file_path = os.path.join(root_path, file_name)
        with rasterio.open(file_path) as dataset:
            elevation = dataset.read(1)
        elevation[elevation <= filter] = np.nan
    elevation, colors = _get_color_normalization(elevation, colors)
    cmap = LinearSegmentedColormap.from_list('custom_cmap', colors)
    # 添加各种特征
    land = cfeature.NaturalEarthFeature('physical', 'land', '50m', edgecolor='black', facecolor='none',
                                        linewidth=0.4)
    ax.add_feature(land)
    ax.imshow(elevation, origin='upper', extent=extents, transform=proj, cmap=cmap)
    ax.add_feature(cfeature.OCEAN)
    ax.add_feature(cfeature.LAND)
    # ax.add_feature(cfeature.LAKES, edgecolor='black')
    ax.add_feature(cfeature.RIVERS)
    # ax.add_feature(cfeature.BORDERS)
    # 添加网格线
    # ax.gridlines(linestyle='--')
    # 设置大刻度和小刻度
    tick_proj = ccrs.PlateCarree()
    ax.set_xticks(np.arange(area[0][0], area[0][1] + 1, area[0][2]), crs=tick_proj)
    # ax.set_xticks(np.arange(-180, 180 + 30, 30), minor=True, crs=tick_proj)
    ax.set_yticks(np.arange(area[1][0], area[1][1] + 1, area[1][2]), crs=tick_proj)
    # ax.set_yticks(np.arange(-90, 90 + 15, 15), minor=True, crs=tick_proj)
    # 利用Formatter格式化刻度标签
    ax.xaxis.set_major_formatter(LongitudeFormatter())
    ax.yaxis.set_major_formatter(LatitudeFormatter())
    if out_folder is None:
        if type(file_name) == list:
            file_name = file_name[0]
            file_dir = 'concate'
        else:
            file_dir = 'outs'
    else:
        file_dir = out_folder
    os.makedirs(os.path.join(root_path, file_dir), exist_ok=True)
    file_name = file_name.replace('.tif', '.png')
    file_title = datetime.strptime(file_name.split('-')[0].replace('target_', ''), '%Y%m%d_%H%M%S')
    file_name = file_title.strftime('%Y%m%d_%H%M%S.png')
    exposure = exposure if exposure else (60 if file_dir == 'concate' else 15)
    if tz == 'BJT':
        exposure_end = (file_title + timedelta(hours=8) + timedelta(minutes=exposure)).strftime('%H:%M')
        file_title = (file_title + timedelta(hours=8)).strftime('%Y-%m-%d %H:%M')
    else:
        exposure_end = (file_title + timedelta(minutes=exposure)).strftime('%H:%M')
        file_title = file_title.strftime('%Y-%m-%d %H:%M')
    ax.set_title(file_title + f'-' + exposure_end + f' ({tz})', fontsize=font_size)
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    # plt.title(fontsize=font_size)
    file_save_path = os.path.join(root_path, file_dir, file_name)
    plt.savefig(file_save_path)
    if zoom_rectangle is not None:
        read_png = cv2.imread(file_save_path)
        read_png = zoomAndDock(read_png, zoom_rectangle, docker, scale_factor=5, border=14)
        cv2.imwrite(file_save_path, read_png)
    crop_png(file_save_path, left=corp[0], top=corp[1], right=corp[2], bottom=corp[3])
    return file_save_path
    # plt.show()



import os
from datetime import datetime, timedelta
import netCDF4 as nc
import numpy as np
from einops import rearrange, repeat
from jacksung.utils.data_convert import np2tif, get_transform_from_lonlat_matrices, Coordinate
import xarray as xr
from pyresample import create_area_def, kd_tree
from pyresample.geometry import AreaDefinition
from jacksung.utils.time import Stopwatch
from jacksung.utils.cache import Cache


def get_resample_infos(hdf_path, lock=None, cache=None):
    if lock:
        lock.acquire()
    ds = nc.Dataset(hdf_path)
    if lock:
        lock.release()
    ld = float(ds['nominal_satellite_subpoint_lon'][:])
    ld = round(ld, 2)
    if cache:
        cache_result = cache.get_key_in_cache(ld)
        if cache_result is not None:
            return cache_result
    # 原始GEOS投影
    goes_proj_str = ds['goes_imager_projection']
    h = float(goes_proj_str.perspective_point_height)
    x = ds['x'][:] * h  # 投影x坐标 (radians)
    y = ds['y'][:] * h  # 投影y坐标 (radians)
    # 计算投影范围 (根据x, y的边界)
    half_pixel_width = (x[1] - x[0]) / 2.0
    half_pixel_height = (y[1] - y[0]) / 2.0
    area_extent = (x[0] - half_pixel_width, y[-1] - half_pixel_height,
                   x[-1] + half_pixel_width, y[0] + half_pixel_height)
    goes_area = AreaDefinition(
        area_id='goes_fixed_grid', proj_id='goes_geos', description='GOES Fixed Grid',
        projection={
            'proj': 'geos',
            'lon_0': goes_proj_str.longitude_of_projection_origin,
            'h': goes_proj_str.perspective_point_height,
            'x_0': 0,
            'y_0': 0,
            'a': goes_proj_str.semi_major_axis,
            'b': goes_proj_str.semi_minor_axis,
            'sweep': goes_proj_str.sweep_angle_axis
        },

        width=len(x), height=len(y), area_extent=area_extent)
    left = ld - 60
    right = ld + 60
    target_areas = []
    if left < -180:
        # 跨越180度经线，分两部分重采样
        # 左半部分
        target_area_left = create_area_def(
            area_id='wgs84_left', projection='EPSG:4326', area_extent=[left + 360, -60, 180, 60],
            resolution=(0.05, 0.05), units='degrees')
        target_areas.append(target_area_left)
        # 右半部分
        target_area_right = create_area_def(
            area_id='wgs84_right', projection='EPSG:4326', area_extent=[-180, -60, right, 60],
            resolution=(0.05, 0.05), units='degrees')
        target_areas.append(target_area_right)
    else:
        target_area = create_area_def(
            area_id='wgs84', projection='EPSG:4326', area_extent=[left, -60, right, 60],
            resolution=(0.05, 0.05), units='degrees')
        target_areas.append(target_area)
    resample_infos = []
    for target_area in target_areas:
        # 使用最近邻法重采样，对于分类数据；对于连续数据，可以使用 ‘bilinear’
        resample_infos.append(
            kd_tree.get_neighbour_info(goes_area, target_area, radius_of_influence=5000, neighbours=1))
    if cache:
        cache.add_key(ld, resample_infos)
    return resample_infos


def getSingleChannelNPfromHDF(hdf_path, lock=None, return_coord=False, only_coord=False, resample_infos=None):
    if lock:
        lock.acquire()
    ds = nc.Dataset(hdf_path)
    if lock:
        lock.release()
    ld = float(ds['nominal_satellite_subpoint_lon'][:])
    np_data = np.array(ds['Rad'][:]).astype(np.float32)
    ld = round(ld, 2)
    left = ld - 60
    right = ld + 60
    coord = Coordinate(left=left, bottom=-60, right=right, top=60, x_res=0.05, y_res=0.05)
    if only_coord:
        return coord
    np_datas = []
    if resample_infos is None:
        resample_infos = get_resample_infos(hdf_path, lock=lock)
    for info in resample_infos:
        valid_input_index, valid_output_index, index_array, distance_array = info
        # 使用最近邻法重采样，对于分类数据；对于连续数据，可以使用 ‘bilinear’
        results = kd_tree.get_sample_from_neighbour_info(
            'nn', output_shape=(coord.h, int(len(info[1]) / 2400)), data=np_data, valid_input_index=valid_input_index,
            valid_output_index=valid_output_index, index_array=index_array, fill_value=np.nan)
        np_datas.append(results)
    # 合并两部分数据
    np_data = np.concatenate(np_datas, axis=1)
    if return_coord:
        return np_data, coord
    else:
        return np_data


def get_filename_by_date_from_dir(dir_path, date, satellite='G18'):
    file_lists = {}
    for file in os.listdir(dir_path):
        if not file.endswith('.nc'):
            continue
        splits = file.split('_')
        year = int(splits[3][1:5])
        doy = int(splits[3][5:8])
        hour = int(splits[3][8:10])
        minute = int(splits[3][10:12])
        file_date = datetime(year=year, month=1, day=1) + timedelta(days=doy - 1, hours=hour, minutes=minute)
        if date == file_date and splits[2] == satellite:
            file_lists[int(splits[1].split('-')[3][3:])] = file
    return file_lists


def getNPfromDir(dir_path, date, satellite='G18', lock=None, return_coord=False, infos=None, cache=None):
    np_data = None
    coord = None
    data_channel_count = 0
    files = get_filename_by_date_from_dir(dir_path, date, satellite)
    for channel, file in files.items():
        if infos is None:
            infos = get_resample_infos(os.path.join(dir_path, file), lock=lock, cache=cache)
        channel_data, coord = getSingleChannelNPfromHDF(
            os.path.join(dir_path, file), return_coord=True, resample_infos=infos)
        if channel_data is None:
            raise Exception(f"文件{file}，通道 {channel} 数据获取失败")
        if np_data is None:
            np_data = np.full([9] + list(channel_data.shape), np.nan)
        np_data[channel - 8] = channel_data
        data_channel_count += 1
    if data_channel_count < 9:
        raise Exception(
            f"文件夹{dir_path}中，卫星 {satellite} 在时间 {date} 的数据通道不完整，仅获取到 {data_channel_count} 个通道")
    if return_coord:
        return np_data, coord
    else:
        return np_data


if __name__ == '__main__':
    np_data = getNPfromDir(rf'D:\python_Project\Huayu_Global\file_download\2022\12\30',
                           datetime(year=2022, month=12, day=30, hour=3, minute=0), satellite='G18')
    np2tif(np_data, save_path='test_goes', out_name='GOES18',
           left=-137 - 60, top=60, x_res=0.05, y_res=0.05, dtype=np.float32)
    np_data = getNPfromDir(rf'D:\python_Project\Huayu_Global\file_download\2022\12\30',
                           datetime(year=2022, month=12, day=30, hour=3, minute=0), satellite='G16')
    np2tif(np_data, save_path='test_goes', out_name='GOES16',
           left=-75.2 - 60, top=60, x_res=0.05, y_res=0.05, dtype=np.float32)



import sys
import os
import re
import subprocess
import select
import argparse
from termcolor import colored

MEMORY_FREE_RATIO = 0.05
MEMORY_MODERATE_RATIO = 0.9
GPU_FREE_RATIO = 0.05
GPU_MODERATE_RATIO = 0.75


def colorize(_lines):
    for i in range(len(_lines)):
        line = _lines[i]
        m = re.match(r"\| (?:N/A|..%)\s+[0-9]{2,3}C.*\s([0-9]+)MiB\s+\/\s+([0-9]+)MiB.*\s([0-9]+)%", line)
        if m is not None:
            used_mem = int(m.group(1))
            total_mem = int(m.group(2))
            gpu_util = int(m.group(3)) / 100.0
            mem_util = used_mem / float(total_mem)

            is_low = is_moderate = is_high = False
            is_high = gpu_util >= GPU_MODERATE_RATIO or mem_util >= MEMORY_MODERATE_RATIO
            if not is_high:
                is_moderate = gpu_util >= GPU_FREE_RATIO or mem_util >= MEMORY_FREE_RATIO

            if not is_high and not is_moderate:
                is_free = True

            c = 'red' if is_high else ('yellow' if is_moderate else 'green')
            _lines[i] = colored(_lines[i], c)
            _lines[i - 1] = colored(_lines[i - 1], c)

    return _lines


def main():
    parser = argparse.ArgumentParser(
        prog='nvidia-watch',  # 程序名
        description='watch gpu',  # 描述
        epilog='Copyright(r), 2023'  # 说明信息
    )
    parser.add_argument('-l', '--command-length', default=20, const=100, type=int, nargs='?')
    parser.add_argument('-c', '--color', action='store_true')
    args = parser.parse_args()

    # parse the command length argument
    command_length = args.command_length
    color = args.color
    processes = subprocess.run('nvidia-smi', stdout=subprocess.PIPE)
    lines = processes.stdout.decode().split("\n")[:-1]
    lines_to_print = []
    # Copy the utilization upper part verbatim
    pid_idx = 0
    for i in range(len(lines)):
        if not lines[i].startswith("| Processes:"):
            # if lines[i].count('MIG M.') > 0 or lines[i].count('N/A') > 0:
            #     continue
            lines_to_print.append(lines[i].rstrip())
        else:
            pid_idx = i + 3
            break

    if color:
        lines_to_print = colorize(lines_to_print)

    for line in lines_to_print:
        print(line)

    # Parse the PIDs from the lower part
    gpu_num = []
    pid = []
    gpu_mem = []
    user = []
    cpu = []
    mem = []
    time = []
    command = []
    for i in range(pid_idx, len(lines)):
        if lines[i].startswith("+--") or lines[i].startswith('|==') or "Not Supported" in lines[i]:
            continue
        no_running_process = "No running processes found"
        if no_running_process in lines[i]:
            print("|  " + no_running_process + " " * (83 - len(no_running_process)) + "  |")
            print(lines[-1])
            sys.exit()
        line = lines[i]
        line = re.split(r'\s+', line)
        gpu_num.append(line[1])
        pid.append(line[4])
        gpu_mem.append(line[7])
        user.append("")
        cpu.append("")
        mem.append("")
        time.append("")
        command.append("")

    # Query the PIDs using ps
    ps_format = "pid,user,%cpu,%mem,etime,command"
    processes = subprocess.run(["ps", "-o", ps_format, "-p", ",".join(pid)], stdout=subprocess.PIPE)

    # Parse ps output
    for line in processes.stdout.decode().split("\n"):
        if line.strip().startswith('PID') or len(line) == 0:
            continue
        parts = re.split(r'\s+', line.strip(), 5)
        idx = pid.index(parts[0])
        user[idx] = parts[1]
        cpu[idx] = parts[2]
        mem[idx] = parts[3]
        time[idx] = parts[4] if not "-" in parts[4] else parts[4].split("-")[0] + " days"
        command[idx] = parts[5][0:100]

    format = ("| %5s %10s %8s %10s %8s %8s %8s %-" + str(command_length) + "." + str(command_length) + "s  |")

    print(format % (
        "GPU", "PID", "USER", "GPU MEM", "%CPU", "%MEM", "TIME", "COMMAND"
    ))

    for i in range(len(pid)):
        # continue
        print(format % (
            gpu_num[i],
            pid[i],
            user[i],
            gpu_mem[i],
            cpu[i],
            mem[i],
            time[i],
            command[i]
        ))

    print(lines[-1])


if __name__ == '__main__':
    main()
